from ..comm.http import HttpClient
from .base import BaseCallback
from typing import Any


class WebhookCallback(BaseCallback):
    """A Callback for sending an HTTP Post message

    Args:
        url (str): url for webhook callback
        token (str): security bearer token
    """

    def __init__(self, url: str, token: str):
        self.url = url
        self.token = token

    def callback_function(self, data: Any) -> bool:
        http = HttpClient()
        try:
            http.post(self.url, data, self.token)
        except Exception:
            return False
        return True



from marshmallow_dataclass import dataclass
from typing import List, Optional, Any

from ..misc.enum import FaasOpState

from google.cloud.firestore import CollectionReference


@dataclass
class FaasProcess:
    """A Job excution group, usually used for batch processing"""

    jobs_state: str  # Dict[str, FaasOpState]


@dataclass
class FaasJob:
    """State of execution of nested FaaS functions, where each function operation
    execution is called job"""

    name: str
    args: Optional[Any]
    op_id: Optional[str]
    state: FaasOpState
    start_date: int
    end_date: Optional[int]
    total_jobs: int
    ended: Optional[bool]


@dataclass
class FaasError:
    """Error data used to send on an error callback event"""

    job_name: Optional[str]
    date: int
    job_id: Optional[str]
    exception_class: str
    exception_message: str
    exception_file: str
    exception_line: int


@dataclass
class FaasJobTrigger:
    """FaaS job trigger for"""

    name: str
    message: dict
    queue: str
    _job: Optional[FaasJob] = None
    """FaasJob added when the job is triggered"""
    _collection: Optional[CollectionReference] = None
    """Firestore collection reference"""
    _job_id: Optional[str] = None
    """Job id in job subcollection"""



from marshmallow_dataclass import dataclass
from typing import List, Optional

from .enum import CallbackType


@dataclass
class Request:
    """Request base data class"""

    callback_type: Optional[CallbackType]
    callback_location: Optional[str]
    callback_token: Optional[str]
    err_callback_type: Optional[CallbackType]
    err_callback_location: Optional[str]
    err_callback_token: Optional[str]
    job_id: Optional[str]
    op_id: Optional[str]
    job_child_idx_list: Optional[List[int]]


@dataclass
class Response:
    code: int


@dataclass
class ResponseStatus(Response):
    status: str
    message: str


@dataclass
class ResponseError(Response):
    error: str
    code: int


@dataclass
class CallbackCreation:
    type: Optional[CallbackType]
    location: Optional[str]
    token: Optional[str]



from functools import wraps
from typing import Any, Callable

from ..misc.exception import CallbackRunError
from ..misc.singleton import Singleton


class BaseCallback(metaclass=Singleton):
    def add_callback(self, function: Callable):
        """Decorate function to add a callback at the end of its execution

        Args:
            function (Callable): decorated function

        Raises:
            CallbackRunError: throws error if callback is not run

        Returns:
            Any: function result
        """

        @wraps(function)
        def wrapper(*args, **kwargs):
            result = function(*args, **kwargs)
            if self.callback_function(result):
                return result
            else:
                raise CallbackRunError()

        return wrapper

    def callback_function(self, data: Any) -> bool:
        """run callback logic given an input data

        Args:
            data (Any): input for callback

        Raises:
            NotImplementedError: if it is not implemented

        Returns:
            bool: True if success
        """

        raise NotImplementedError("not implemented")



from typing import Dict, Optional, Type
from .base import BaseCallback
from .pubsub import PubSubCallback
from .webhook import WebhookCallback
from ..misc.enum import CallbackType
from ..misc.model import CallbackCreation, Request

import re


__callback_type_dict: Dict[CallbackType, Type[BaseCallback]] = {
    CallbackType.WEBHOOK: WebhookCallback,
    CallbackType.PUBSUB: PubSubCallback,
}


def _get_callback_data(req: Request) -> Optional[CallbackCreation]:
    if req.callback_type:
        return CallbackCreation(
            type=req.callback_type,
            token=req.callback_token,
            location=req.callback_location,
        )
    else:
        return None


def _get_err_callback_data(req: Request) -> Optional[CallbackCreation]:
    if req.callback_type:
        return CallbackCreation(
            type=req.err_callback_type,
            token=req.err_callback_token,
            location=req.err_callback_location,
        )
    else:
        return None


def _get_callback(callback_data: CallbackCreation) -> Optional[BaseCallback]:
    if callback_data is not None:
        callback_class = __callback_type_dict[callback_data.type]
        callback_instance: BaseCallback
        if WebhookCallback == callback_class:
            url = callback_data.location
            token = callback_data.token
            callback_instance = WebhookCallback(url, token)
        if PubSubCallback == callback_class:
            project = re.findall(r"projects/(.*?)/", callback_data.location)
            topic = re.findall(r".*topics/(.*)", callback_data.location)
            callback_instance = PubSubCallback(project, topic)
        return callback_instance
    else:
        return None


def get_callback(req: Request) -> Optional[BaseCallback]:
    callback_data = _get_callback_data(req)
    return _get_callback(callback_data)


def get_err_callback(req: Request) -> Optional[BaseCallback]:
    callback_data = _get_err_callback_data(req)
    return _get_callback(callback_data)



from ..comm.pubsub import PubSubClient
from .base import BaseCallback
from typing import Any
import json


class PubSubCallback(BaseCallback):
    """A Callback for sending a message to the PubSub Queue Service

    Args:
        project_id (str): Google Cloud project id
        topic_name (str): PubSub topic name
    """

    def __init__(self, project_id: str, topic_name: str):
        self.project = project_id
        self.topic = topic_name

    def callback_function(self, data: Any) -> bool:
        pubsub = PubSubClient(self.project, self.topic)
        try:
            pubsub.send_message(json.dumps(data))
        except Exception:
            return False
        return True
